package eve.ui;
import eve.sys.Handle;
import eve.sys.Task;
/**
* A ProgressAndControl panel allows a ProgressBar to be displayed in the same
* location as another Control. When a task starts, the ProgressBar can be displayed
* and the other Control hidden. When the task ends, the other Control will be 
* redisplayed.
**/
//##################################################################
public class ProgressAndControl extends CardPanel implements ProgressDisplay{
//##################################################################
/**
* This is the ProgressBarForm. You can alter its parameters as you wish.
**/
public ProgressBarForm pbf;
/**
* Add your controls to this panel.
**/
public CellPanel controls;

//===================================================================
public ProgressAndControl(ProgressBarForm form)
//===================================================================
{
	pbf = form;
	addItem(controls = new CellPanel(),"controls",null);
	addItem(pbf,"progress",null);
}
//===================================================================
public ProgressAndControl()
//===================================================================
{
	pbf = new ProgressBarForm();
	pbf.setBorder(EDGE_ETCHED,2);
	pbf.showMainTask = true; pbf.showSubTask = false;
	pbf.showTaskInBar = true; pbf.showStop = true;
	pbf.horizontalLayout = true;
	pbf.exitOnCompletion = false;
	pbf.dontExitOnStopPressed = true;
	addItem(controls = new CellPanel(),"controls",null);
	addItem(pbf,"progress",null);
}
/**
* Call this when the task begins.
**/
//===================================================================
public Handle startTask(Handle h,String task)
//===================================================================
{
	pbf.setTask(h,task);
	select("progress");
	return h;
}
/**
* Call this when you consider the task complete and want to show
* the controls again.
**/
//===================================================================
public void endTask()
//===================================================================
{
	select("controls");
}
/**
 * This is a safe way to execute a Task with ProgressAndControl within
 * an Event thread. It uses newEventThread() to start a new Event thread
 * while it waits for the task to complete.
 * @param h the handle of the Task to display on the ProgressBar
 * @param name a title for the Task that is running.
 * @param disableWhileWorking an optional Control that will be disabled during
 * the task.
 * @return the status of the task when completed.
 */
public int executeTask(Handle h, String name, Control disableWhileWorking)
{
	Object obj = newEventThread(disableWhileWorking);
	try{
		startTask(h,name);
		h.waitUntilCompletion();
	}finally{
		endTask();
		resumeEventThread(obj);
	}
	return h.check();
}
/**
 * This method will cause the ProgressAndControl to switch to the progress display,
 * and optionally disable a Control during the task. When the task is complete it will
 * switch back to the Control and re-enable the disabled control.
 * @param h the handle of the Task to display on the ProgressBar
 * @param name a title for the Task that is running.
 * @param disableWhileWorking an optional Control that will be disabled during
 * the task.
 * @return a new Handle that monitors the state of the ProgressAndControl. It will
 * only be stopped <b>after</b> the display has switched back to the ProgressAndControl.
 */
public Handle doTask(final Handle h, String name, Control disableWhileWorking)
{
	final Control c =  disableWhileWorking;
	int oc = 0;
	if (c != null){
		oc = c.modify(Disabled,0);
		c.repaintNow();
	}
	final int old = oc;
	startTask(h,name); 
	return new Task(){
		protected void doRun(){
			h.waitUntilCompletion();
			endTask();
			if (c != null){
				//synchronized(guiLock){
					c.restore(old,Disabled);
					c.repaintNow();
				//}
			}
		}
	}.start();
}

//##################################################################
/* (non-Javadoc)
 * @see eve.ui.ProgressDisplay#setTask(eve.sys.Handle, java.lang.String)
 */
public void setTask(Handle h, String taskName) {
	startTask(h,taskName);
}
/* (non-Javadoc)
 * @see eve.ui.ProgressDisplay#clearTask()
 */
public void clearTask() {
	endTask();
}
}
//##################################################################

