package eve.util;
/**
* A ByteArray is a utility class that represents a number of data bytes. 
* It is generally used as a buffer for data storage.<p>
* Associated with the ByteArray is a byte [] value called "data" which holds
* the data bytes (starting from index 0) and the "length" variable specifies how
* many of those bytes are considered valid (which may be less than the actual
* length of the data variable).
**/
//##################################################################
public class ByteArray implements Copyable, Encodable{
//##################################################################
/**
* The data bytes.
**/
public byte [] data = new byte[0]; //Do not move this variable! It must be first.
/**
* The number of valid bytes in the data.
**/
public int length = 0;//Do not move this variable! It must be second.
/**
* Get a copy of this ByteArray with its own copy of this data.
**/
//===================================================================
public Object getCopy()
//===================================================================
{
	ByteArray ba = new ByteArray();
	ba.length = length;
	ba.data = new byte[length];
	if (length != 0) System.arraycopy(data,0,ba.data,0,length);
	return ba;
}
/**
* This adds space to the ByteArray at the specified index, increasing the 
* length value by numBytes. After this is
* done you can safely write into the data starting at index "where" for
* a length of numBytes.
**/
//===================================================================
public boolean makeSpace(int where,int numBytes)
//===================================================================
{
	if (numBytes <= 0) return true;
	if (length+numBytes > data.length) {
		int expansion = data.length;
		if (expansion > 100000) expansion = 100000;
		if ((length+numBytes) > (expansion+data.length))
			expansion = 1+length+numBytes-data.length;
		int newLen = data.length+expansion;
		//if (true) newLen = length+numBytes;
		byte [] nd = new byte[newLen];
		System.arraycopy(data,0,nd,0,length);
		data = nd;
	}
	if (where < length)
		System.arraycopy(data,where,data,where+numBytes,length-where);
	length += numBytes;
	return true;
}
/**
* This writes into the data at the specific location. Make sure you have called
* makeSpace() before if necessary.
**/
//===================================================================
public void write(int where,byte [] src,int start,int length)
//===================================================================
{
	System.arraycopy(src,start,data,where,length);
}
/**
* Insert bytes into the ByteArray.
 * @param where The index in this ByteArray where the data will go.
 * @param src The source data bytes.
 * @param start The start index in the source data.
 * @param length The number of bytes to copy.
 */
//===================================================================
public void insert(int where, byte[] src,int start,int length)
//===================================================================
{
	if (makeSpace(where,length)) write(where,src,start,length);
}
/**
* Add bytes to the end of the ByteArray.
 * @param src The source data bytes.
 * @param start The start index in the source data.
 * @param length The number of bytes to copy.
 */
//===================================================================
public void append(byte [] src,int start,int length)
//===================================================================
{
	insert(this.length,src,start,length);
}
static byte [] single;
/**
 * Append a single byte of data.
 * @param data The byte to append.
 */
//===================================================================
public void append(byte data)
//===================================================================
{
	if (single == null) single = new byte[1];
	single[0] = data;
	append(single,0,1);
}
/**
* Insert an integer value as up to four bytes in big-endian format.
* @param where the index at which to insert the integer.
* @param value the value to insert.
* @param numBytes the number of bytes (starting from the least significant) to insert.
*/
//===================================================================
public void insertInt(int where,int value,int numBytes)
//===================================================================
{
	makeSpace(where,numBytes);
	Utils.writeInt(value,data,where,numBytes);
}
/**
* Append an integer value as up to four bytes in big-endian format.
* @param value the value to append.
* @param numBytes the number of bytes (starting from the least significant) to insert.
*/
//===================================================================
public void appendInt(int value,int numBytes)
//===================================================================
{
	insertInt(length,value,numBytes);
}
/**
* Append a four-byte integer value.
* @param value the value to append.
*/
//===================================================================
public void appendInt(int value)
//===================================================================
{
	insertInt(length,value,4);
}
/**
* Insert a long value as eight bytes in big-endian format.
* @param where the index at which to insert the integer.
* @param value the value to insert.
*/
//===================================================================
public void insertLong(int where,long value)
//===================================================================
{
	makeSpace(where,8);
	Utils.writeLong(value,data,where);
}
/**
* Append a long value as eight bytes in big-endian format.
* @param value the value to append.
*/
//===================================================================
public void appendLong(long value)
//===================================================================
{
	insertLong(length,value);
}
/**
* This removes bytes from within the data.
**/
//===================================================================
public boolean delete(int where,int numBytes)
//===================================================================
{
	if (numBytes <= 0) return true;
	if (where+numBytes > length) numBytes = length-where;
	if (numBytes > 0){
		System.arraycopy(data,where+numBytes,data,where,length-(where+numBytes));
		length -= numBytes;
	}
	return true;
}
/**
* This sets the length of the valid data within the array, but does NOT extend OR truncate
* the array. If the supplied len is greater than the length of the byte array, the value
* will be reduced to the length of the byte array.
**/
//===================================================================
public void setLength(int len) {if (len > data.length) len = data.length; if (len < 0) len = 0; length = len;}
//===================================================================
/**
* Return a copy of the data in the ByteArray as a byte array of the exact length of the data.
**/
//===================================================================
public byte [] toBytes()
//===================================================================
{
	byte [] got = new byte[length];
	System.arraycopy(data,0,got,0,length);
	return got;
}
//===================================================================
public ByteArray() {}
//===================================================================
/**
* This will actually use the "data" byte array as the data for the byte array. It will not copy the data.
**/
//===================================================================
public ByteArray(byte [] data)
//===================================================================
{
	this.data = data;
	this.length = data.length;
}
/**
Copy the bytes from the supplied array, re-using the interal byte array if possible.
**/
//===================================================================
public void copyFrom(byte[] data, int offset, int length)
//===================================================================
{
	this.length = 0;
	if (data != null) append(data,offset,length);
}
/**
Copy the bytes from the supplied array, re-using the interal byte array if possible.
**/
//===================================================================
public void copyFrom(byte[] data)
//===================================================================
{
	this.length = 0;
	if (data != null) append(data,0,data.length);
}
public void setData(byte[] data)
{
	this.data = data;
	this.length = data.length;
}
/**
 * Set the "length" field to zero and return this CharArray object.
 */
public ByteArray clear()
{
	length = 0;
	return this;
}
/**
* This will ensure that the size of the "data" byte array is at least
* newCapacity. If the size is already greater than or equal to newCapacity, then
* this will have no effect.
* @param newCapacity The minimum size of the data array.
*/
//===================================================================
public void ensureCapacity(int newCapacity)
//===================================================================
{
	if (data != null && data.length >= newCapacity) return;
	int sz = data == null ? 0 : data.length;
	if (newCapacity < (sz*2+2)) newCapacity = sz*2+2;
	toNewLength(newCapacity);
}
/**
* Moves the data into a new data array of the specified length - which
* must be greater than or equal to the number of elements in the array.
**/
//-------------------------------------------------------------------
protected void toNewLength(int newLength)
//-------------------------------------------------------------------
{	
	byte[] nd = new byte[newLength];
	if (length != 0) System.arraycopy(data,0,nd,0,length);
	data = nd;
}
public static String toString(byte[] data, boolean putBrackets)
{
	return toString(data,0,data.length,putBrackets);
}
public static String toString(byte[] data, int offset, int length,boolean putBrackets)
{
	char[] ch = new char[length+(putBrackets ? 2 : 0)];
	int off = putBrackets ? 1 : 0;
	for (int i = 0; i<length; i++){
		char c = (char)(data[i+offset] & 0xff);
		if (c < 20 || c >= 127) c = '.';
		ch[i+off] = c;
	}
	if (putBrackets){
		ch[0] = '['; ch[length+1] = ']';
	}
	return new String(ch);
}
public String toString()
{
	return toString(data,0,length,true);
}
public int getChecksum()
{
	int cs = 0;
	for (int i = 0; i<length; i++)
		cs += (int)(data[i]) & 0xff;
	return cs;
}
//##################################################################
}
//##################################################################


