/*
 * Created on Mar 25, 2006
 *
 * Michael L Brereton - www.ewesoft.com
 * 
 * 
 */
package eve.util;

import java.util.Vector;

import eve.sys.Cache;
import eve.sys.Device;
import eve.sys.Vm;
import eve.sys.WeakRef;

/**
 * A StringList is used to provide a Vector/array of String/SubString values from
 * String, SubString, CharArray or char[] initial values in a way that minimizes
 * Object creation by reusing char[] buffers and SubString values as is possible.
 * <p>
 * It can also be used to split Strings into re-used SubString values. 
 * This class is <b>not</b> synchronized.
 * 
 * @author Michael L Brereton
 *
 */
//####################################################
public class StringList{
	//
	//private Vector added = new Vector();
	private int count;
	private int[] lengths, offsets;
	private Object[] added;
	private CharArray buffer;
	private CharArray userBuffer;
	//
	private Object[] array;
	private Vector made;
	//
	private SubString subString = new SubString();
	private static final int STATE_ADDING = 0;
	private static final int STATE_STRINGS = 1;
	private static final int STATE_SUBSTRINGS = 2;
	private int state;
	private int numNull;
	//private int numStrings, numSubStrings;
	private WeakRef usedSubs = new WeakRef(true); // This is a Vector.
	//private int addedSubStrings;
	public int expandSize = 10;
	//{new Exception().printStackTrace();}
	private static Vector splitVector = new Vector();
	
	/**
	 * Return the number of individual Strings.
	 * @return
	 */
	public int size()
	{
		return count;
	}
	public void ensureCapacity(int size)
	{
		int curSize = lengths == null ? 0 : lengths.length;
		if (curSize >= size) return;
		if (size < curSize+expandSize) size = curSize+expandSize;
		if (lengths == null || lengths.length < size){
			int[] nl = new int[size];
			if (count != 0) System.arraycopy(lengths,0,nl,0,count);
			lengths = nl;
			nl = new int[size];
			if (count != 0) System.arraycopy(offsets,0,nl,0,count);
			offsets = nl;
			Object[] nv = new Object[size];
			if (count != 0) System.arraycopy(added,0,nv,0,count);
			added = nv;
		}
	}
	private void reclaim()
	{
		if (state == STATE_SUBSTRINGS && buffer != null){ //There are substrings in there that I added.
			char[] data = buffer.data;
			Vector used = (Vector)usedSubs.get();
			if (used == null) used = new Vector();
			used.ensureCapacity(count);
			for (int i = 0; i<count; i++){
				Object obj = added[i];
				if (obj instanceof SubString && ((SubString)obj).data == data)
					used.add(obj);
			}
			usedSubs.set(used);
		}
		if (added != null) Utils.zeroArrayRegion(added,0,count);
		if (array != null) Utils.zeroArrayRegion(array,0,array.length);
		if (made != null) made.removeAllElements();
		state = STATE_ADDING;
	}

	/**
	 * Create an empty string list.
	 */
	public StringList()
	{
		
	}
	/**
	 * Create a new StringList and add text to it.
	 * @param text the text to add.
	 */
	public StringList(String text)
	{
		this();
		add(text);
	}
	/**
	 * Create a new StringList and split and add text to it.
	 * @param text the text to split.
	 * @param separator the separator character.
	 */
	public StringList(String text, char separator)
	{
		this();
		split(text,separator);
	}
	/**
	 * This sets the number of items in the StringList to zero and
	 * caches all the used SubStrings for future use.
	 */
	public StringList zero()
	{
		reclaim();
		numNull = count = 0;
		if (buffer != null) buffer.length = 0;
		return this;
	}
	public StringList clear()
	{
		zero();
		added = null;
		lengths = null;
		offsets = null;
		buffer = null;
		made = null;
		array = null;
		usedSubs.clear();
		userBuffer = null;
		return this;
	}
	public CharArray getUserBuffer()
	{
		if (userBuffer == null) userBuffer = new CharArray();
		userBuffer.clear();
		return userBuffer;
	}
	public StringList setSplitter(SubString splitter)
	{
		if (!subString.getClass().equals(splitter.getClass()))
				usedSubs.clear();
		subString = splitter;
		return this;
	}
	public SubString getSplitter()
	{
		return subString;
	}
	public StringList setSplitter(Class splitterClass)
	{
		if (splitterClass == null) splitterClass = SubString.class;
		if (!subString.getClass().equals(splitterClass))
			setSplitter((SubString)Cache.get(splitterClass));
		return this;
	}
	/*
	private void add(char[] data, int start, int length)
	{
		if (buffer == null) buffer = new CharArray();
		int where = buffer.length;
		buffer.append(data,start,length);
		if (used == null) used = (Vector)usedSubs.get();
		int sz = used == null ? 0 : used.size();
		SubString ss = null;
		if (sz != 0) {
			ss = (SubString)used.get(sz-1);
			used.removeElementAt(sz-1);
		}
		if (ss == null) ss = (SubString)subString.getNew();
		ss.set(buffer.data,where,length);
	}
	*/
	public void addNull()
	{
		add(null,0,0,null);
	}
	public SubString add(SubString str)
	{
		if (str == null){
			add(null,0,0,null);
			return str;
		}
		add(str.data,str.start,str.length,null);
		return str;
	}
	public String add(String str)
	{
		if (str == null){
			add(null,0,0,null);
			return str;
		}
		add(Vm.getStringChars(str),0,str.length(),str);
		return str;
	}
	public CharArray add(CharArray ca)
	{
		if (ca == null){
			add(null,0,0,null);
			return ca;
		}
		add(ca.data,0,ca.length,null);
		return ca;
	}
	/**
	 * Try to add a text object. This can be one of the following types:<p>
	 * String, SubString, CharArray, char[], String[], SubString[]
	 * @param obj a text object of the types specified above.
	 * @return true if the type was correct, false if not.
	 */
	public boolean tryAdd(Object obj)
	{
		if (obj instanceof CharArray) add((CharArray)obj);
		else if (obj instanceof SubString) add((SubString)obj);
		else if (obj instanceof char[]) add((char[])obj);
		else if (obj instanceof String) add((String)obj);
		else if (obj instanceof String[]){
			String[] s = (String[])obj;
			for (int i = 0; i<s.length; i++)
				add(s[i]);
		}
		else if (obj instanceof SubString[]){
			String[] s = (String[])obj;
			for (int i = 0; i<s.length; i++)
				add(s[i]);
		}
		else return false;
		return true;
}
	public char[] add(char[] chars)
	{
		return add(chars,0,chars == null ? 0 : chars.length,null);
	}
	public char[] add(char[] chars,int offset,int length)
	{
		return add(chars,offset,length,null);
	}
	private char[] add(char[] chars,int offset,int length,String src)
	{
		if (state != STATE_ADDING) reclaim();
		ensureCapacity(count+1);
		if (chars == null){
			numNull++;
			lengths[count] = -1;
			added[count] = null;
		}else{
			lengths[count] = length;
			if (buffer == null) buffer = new CharArray();
			buffer.append(chars,offset,length);
			//System.out.println("Buffer: "+buffer);
			added[count] = src;
		}
		int off = 0;
		if (count != 0){
			off = offsets[count-1]+lengths[count-1];
			if (off < offsets[count-1])
				off++;
		}
		offsets[count] = off;
		count++;
		return chars;
	}
	/**
	 * This splits text into SubStrings placed into this StringList, after clearing
	 * the list initially. 
	 * @param source
	 * @param start
	 * @param length
	 * @param separator
	 */
	public StringList split(char[] source,int start, int length, char separator)
	{
		zero();
		//if (source != null) System.out.println("Splitting: "+new String(source,start,length));
		Vector used = (Vector)usedSubs.get();
		if (used == null) used = new Vector();
		subString.set(source,start,length);
		if (made == null) made = new Vector();
		Vector splitVector = made;
		splitVector.removeAllElements();
		int sz = subString.split(separator,splitVector,0,used);
		subString.data = null;
		//System.out.println("Split to: "+sz);
		for (int i = 0; i<sz; i++){
			SubString ss = (SubString)splitVector.get(i);
			add(ss);
			used.add(ss);
		}
		splitVector.removeAllElements();
		return this;
	}
	/**
	 * This splits text into SubStrings placed into this StringList, after clearing
	 * the list initially. 
	 * @param source
	 * @param start
	 * @param length
	 * @param separator
	 */
	public StringList split(String source, char separator)
	{
		return split(Vm.getStringChars(source),0,source.length(),separator);
	}
	/**
	 * This splits text into SubStrings placed into this StringList, after clearing
	 * the list initially. 
	 * @param source
	 * @param start
	 * @param length
	 * @param separator
	 */
	public StringList split(String source, int start, int length, char separator)
	{
		return split(Vm.getStringChars(source),start,length,separator);
	}
	
	//private static final int CONVERT_COUNT_NEED = 0;
	private static final int CONVERT_TO_SUBSTRINGS = 1;
	private static final int CONVERT_TO_STRINGS = 2;
	
	private static boolean hasNative = true;
	//private static native Object nativeConvert(int op,Vector added,int toConvert,char[] buffer,Vector used);
	
	private static void convert(int op,int count,Object[] added,int[] lengths,char[] data, Vector used)
	{
		// Now convert all Strings to SubStrings, discarding the original Strings.
		if (hasNative && false)try{
			//return nativeConvert(op,added,toConvert,buffer,used);
		}catch(Throwable t){
			Device.checkNoNativeMethod(t);
			hasNative = false;
		}
		int sz = count;
		/*
		if (op == CONVERT_MAKE_SPACE){
			int need = 0;
			for (int i = 0; i<sz && toDo < toConvert; i++){
				Object obj = added.get(i);
				if (obj instanceof SubString) continue;
				toDo++;
				if (obj instanceof String){
					need += ((String)obj).length();
				}else if (obj instanceof CharArray){
					need += ((CharArray)obj).length;
				}else if (obj instanceof char[]){
					need += ((char[])obj).length;
				}
			}
			if (buffer == null || buffer.length < need)
				buffer = new char[need];
			return buffer;
		}else*/ 
		int idx = 0;
		int usz = used == null ? 0 : used.size();
		for (int i = 0; i<sz; i++){
			int len = lengths[i];
			if (len < 0) continue;
			if (op == CONVERT_TO_SUBSTRINGS){
				usz--;
				SubString ss = ((SubString)used.get(usz)).set(data,idx,len);
				used.removeElementAt(usz);
				added[i] = ss;
			}else{
				if (data != null){
					SubString ss = (added[i] instanceof SubString) ? (SubString)added[i]:null;
					if (ss != null && ss.data == data) used.add(ss);
				}
				if (!(added[i] instanceof String)){
					added[i] = new String(data,idx,len);
				}
			}
			//System.out.println("Did: "+added[i]);
			idx += len;
		}
		/*
		}else{
			for (int i = 0; i<sz && toDo < toConvert; i++){
				Object obj = added.get(i);
				if (obj instanceof SubString){
					toDo++;
					SubString ss = (SubString)obj;
					added.setElementAt(ss.toString(),i);
					if (ss.data == buffer) used.add(ss);
				}else if (obj instanceof CharArray){
					toDo++;
					added.setElementAt(obj.toString(),i);
				}else if (obj instanceof char[]){
					toDo++;
					added.setElementAt(new String((char[])obj),i);
				}
			}
			return added;
		}
		*/
	}
	private void convertToSubStrings()
	{
		if (state == STATE_SUBSTRINGS) return;
		state = STATE_SUBSTRINGS;
		int toConvert = count-numNull;
		if (toConvert <= 0) return;
		Vector used = (Vector)usedSubs.get();
		if (used == null) used = new Vector();
		int usz = used.size();
		if (usz < toConvert){
			used.ensureCapacity(toConvert);
			while(usz < toConvert){
				used.add(subString.getNew());
				usz++;
			}
		}
		convert(CONVERT_TO_SUBSTRINGS,count,added,lengths,buffer == null ? null : buffer.data,used);
		usedSubs.set(used);
	}
	private void convertToPermanentSubStrings()
	{
		reclaim();
		int toConvert = count-numNull;
		if (toConvert <= 0) return;
		if (made == null) made = new Vector();
		made.removeAllElements();
		for (int i = 0; i<toConvert; i++)
			made.add(subString.getNew());
		char[] data = buffer.toChars();
		convert(CONVERT_TO_SUBSTRINGS,count,added,lengths,data,made);
		made.removeAllElements();
	}
	
	private void convertToStrings()
	{
		if (state == STATE_STRINGS) return;
		state = STATE_STRINGS;
		int toConvert = count-numNull;
		if (toConvert <= 0) return;
		Vector used = (Vector)usedSubs.get();
		if (used == null) used = new Vector();
		convert(CONVERT_TO_STRINGS,count,added,lengths,buffer == null ? null : buffer.data,used);
		usedSubs.set(used);
	}
	private Vector toFixed(Vector dest)
	{
		if (dest == null) dest = new Vector();
		dest.ensureCapacity(count);
		for (int i = 0; i<count; i++)
			dest.add(added[i]);
		return dest;
	}
	
	private Vector toMade()
	{
		if (made == null) made = new Vector();
		made.removeAllElements();
		made.ensureCapacity(count);
		for (int i = 0; i<count; i++)
			made.add(added[i]);
		return made;
	}
	/**
	 * Convert all entries into SubStrings. The Vector and SubStrings
	 * within the Vector are temporary only and may be reused by the StringList.
	 * @return a re-used Vector containing the entries as SubString values.
	 */
	public Vector toTempSubStrings()
	{
		convertToSubStrings();
		return toMade();
	}
	/**
	 * Convert all entries into String.  
	 * The Vector is temporary only and may be reused by the StringList.
	 * @return a re-used Vector containing the entries as String values.
	 */
	public Vector toTempStrings()
	{
		convertToStrings();
		return toMade();
	}
	/**
	 * Copy the added Strings as SubStrings into the provided destination Vector.
	 * @param dest a destination Vector to add the SubStrings to. The SubStrings 
	 * placed in the Vector are new and not re-used by the StringList.
	 * @return the destination Vector or a new one if it is null.
	 */
	public Vector toSubStrings(Vector dest)
	{
		convertToPermanentSubStrings();
		dest = toFixed(dest);
		reclaim();
		return dest;
	}
	/**
	 * Copy the added Strings as SubStrings into the provided destination Vector.
	 * @param dest a destination Vector to add the SubStrings to. The SubStrings 
	 * placed in the Vector are not re-used by the StringList.
	 * @return the destination Vector or a new one if it is null.
	 */
	public Vector toStrings(Vector dest)
	{
		convertToStrings();
		return toFixed(dest);
	}
	
	private Object toArray(boolean exactSize, boolean isSubStrings)
	{
		int sz = count;
		if (array != null && (array.length < sz || ((array.length != sz) && exactSize)))
			array = null;
		if (isSubStrings){
			if (!(array instanceof SubString[]))
				array = new SubString[sz];
		}else{
			if (!(array instanceof String[]))
				array = new String[sz];
		}
		if (sz < array.length)
			Utils.zeroArrayRegion(array,sz,array.length-sz);
		//
		if (added != null && array != null)
			System.arraycopy(added,0,array,0,sz);
		//
		return array;
	}
	/**
	 * Returns true if converting to a String array or a Vector of Strings will not
	 * result in any String or SubString creation. This is true if only String
	 * or null values have been added, or a conversion to Strings was most recently
	 * done.
	 * @return true if converting to a String array or a Vector of Strings will not
	 * result in any String or SubString creation.
	 */
	public boolean stringsAreBest()
	{
		if (state == STATE_STRINGS) return true;
		if (state == STATE_SUBSTRINGS) return false;
		int toConvert = count-numNull;
		if (toConvert <= 0) return true;
		for (int i = 0; i<count; i++){
			if (lengths[i] < 0) continue; // The entry is null.
			if (!(added[i] instanceof String)) return false; // A new String must be added.
		}
		return true;
	}
	
	/**
	 * Return either an array of Strings or an array of SubStrings depending on
	 * what stringsAreBest() returns.
	 * @return either an array of Strings or an array of SubStrings depending on
	 * what stringsAreBest() returns.
	 */
	public Object[] toBestText()
	{
		return stringsAreBest() ? (Object[])toStringArray() : (Object[])toSubStringArray();
	}
	/**
	 * Convert all entries into SubStrings. Added String values are converted
	 * into SubStrings after copying their character data into a temporary buffer and
	 * then discarded.
	 * @param exactSize if this is true then the array will be the exact size
	 * as the number of entries, if false then it may be larger, although all
	 * extra elements will be null.
	 * @return a SubString[] containing the entries as SubString values.
	 */
	public SubString[] toTempSubStringArray(boolean exactSize)
	{
		convertToSubStrings();
		return (SubString[])toArray(exactSize,true);
	}
	/**
	 * Convert all entries into String. Added SubString values are converted
	 * to Strings and discarded.
	 * @param exactSize if this is true then the array will be the exact size
	 * as the number of entries, if false then it may be larger, although all
	 * extra elements will be null.
	 * @return a String[] containing the entries as String values.
	 */
	public String[] toTempStringArray(boolean exactSize)
	{
		convertToStrings();
		return (String[])toArray(exactSize,false);
	}
	public String[] toStringArray()
	{
		return toStringArray(new String[count],0,count); 
	}
	public SubString[] toSubStringArray()
	{
		return toSubStringArray(new SubString[count],0,count); 
	}
	/**
	 * Convert all entries into Strings. 
	 * @param dest an optional destination String[].
	 * @param offset the location in the String to put the data.
	 * @param length the number of items to put in the destination.
	 * @return a String[] containing the entries as String values.
	 */
	public String[] toStringArray(String[] dest,int offset,int length)
	{
		convertToStrings();
		if (length > count) length = count;
		if (dest == null) dest = new String[length+offset];
		System.arraycopy(toTempStringArray(false),0,dest,offset,length);
		return dest;
	}
	/**
	 * Convert all entries into SubStrings. 
	 * @param dest an optional destination SubString[].
	 * @return a SubString[] containing the entries as String values. The
	 * returned array and contained SubStrings are not reused by the StringList.
	 */
	public SubString[] toSubStringArray(SubString[] dest,int offset,int length)
	{
		convertToPermanentSubStrings();
		if (length > count) length = count;
		if (dest == null) dest = new SubString[length+offset];
		System.arraycopy(toTempSubStringArray(false),0,dest,offset,length);
		return dest;
	}
	/**
	 * This may return null if the entry at the index is null.
	 * @param index the index of the SubString to retrieve.
	 * @param dest an optional destination SubString.
	 * @return the dest SubString or a new one if dest was null or null if the entry
	 * at the index was null.
	 */
	public SubString get(int index, SubString dest)
	{
		if (lengths[index] == -1) return null;
		if (dest == null) dest = new SubString();
		return dest.set(buffer.data,offsets[index],lengths[index]);
	}
	/*
	public static void main(String[] args)
	{
		StringList sl = new StringList();
		for (int i = 0; i<5; i++){
			int num = Vm.countObjects(false);
			Object obj = sl.split("Hello there, how are you doing today?",' ').toStringArray();
			num = Vm.countObjects(false)-num;
			System.out.println("Created: "+num);
			System.out.println(mString.toString(obj));
		}
	}
	*/
	public static StringList getCached(Class splitterType)
	{
		return ((StringList)Cache.get(StringList.class)).setSplitter(splitterType);
	}
	public static StringList getCached()
	{
		return getCached(SubString.class);
	}
	public StringList cache()
	{
		Cache.put(this);
		Vector v = (Vector)usedSubs.get();
		return null;
	}
	public void cached()
	{
		zero();
	}
}
//####################################################
