package eve.util;
import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;

import eve.sys.WeakSet;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;
/**
This is similar to a Hashtable in that you associate values with keys. However
only weak references are kept to the keys - which means that the key values
can be garbage collected, in which case it will not be possible to access the object
values the keys refer to.<p>
The clean() method explicitly removes entries where the keys have been garbage
collected.<p>
A background thread is created to clean all existing WeakCache objects at regular
intervals. The default interval is 1 minute, but this can be changed by calling
setCleanInterval().
**/
//##################################################################
public class WeakCache{
//##################################################################

	//private static Thread cleanThread;
	//private static long cleanInterval = 60*1000;
static WeakSet caches;
private Hashtable table;
private Vector cleanVector;
ReferenceQueue rq = new ReferenceQueue();

/**
 * This is the time between a key being freed to when a clean() occurs. There
 * is a time delay to allow for other keys to also be freed.
 */
public int cleanDelay = 1000;

//Vector refs = new Vector(), data = new Vector();
/**
 * Constructs a new WeakCache with the default initial capacity
 * and a default load factor of 0.75
 */
//===================================================================
public WeakCache() {
//===================================================================
	this(11, 0.75f);
}
/**
 * Constructs a new WeakCache with the specified initial capacity
 * and a default load factor of 0.75
 */
//===================================================================
public WeakCache(int initialCapacity) {
//===================================================================
	this(initialCapacity, 0.75f);
}

//===================================================================
public WeakCache(int initialCapacity, float loadFactor)
//===================================================================
{
	table = new Hashtable(initialCapacity,loadFactor);
	synchronized(WeakCache.class){
		if (caches == null){
			caches = new WeakSet();
			new Cleaner().start();
		}
		caches.add(this);
	}
	/*
	synchronized(WeakCache.class){
		if (cleanThread == null){
			caches = new WeakSet();
			//cleanVector = new Vector();
			cleanThread = new Thread(){
				public void run(){
					try{
						Vector v = new Vector();
						while(true){
							try{
								try{
									sleep(cleanInterval);
								}catch(Exception e){
								}
								v.removeAllElements();
								int got = caches.getRefs(v);
								int sz = v.size();
								for (int i = 0; i<sz; i++){
									WeakCache wc = (WeakCache)v.get(i);
									if (wc != null) wc.clean();
								}
								v.removeAllElements();
							}catch(Throwable t){}
						}
					}finally{
						synchronized(WeakCache.class){
							if (cleanThread == this) cleanThread = null;
						}
					}
				}
			};
			try{
				cleanThread.setDaemon(true);
			}catch(Throwable t){}
			cleanThread.start();
		}
	}
	caches.add(this);
	*/
}
/**
 * Set the interval between cache cleaning in milliseconds.
 * @param interval the interval in milliseconds. If this is less 1 then the default
 * value of 60000 (1 minute) will be used.
 * @deprecated this is no longer used. Each WeakCache has its own cleaner
 * Thread that checks the reference queue for cleaning.
 */
//===================================================================
public synchronized static void setCleanInterval(long interval)
//===================================================================
{
	if (interval <= 0) interval = 60*1000;
	/*
	cleanInterval = interval;
	if (cleanThread != null) try{
		cleanThread.interrupt();
	}catch(SecurityException e){}
	*/
}

/**
 * Remove all entries for which the key has been garbage collected.
 */
//===================================================================
public void clean()
//===================================================================
{
	Vector v = null;
	synchronized(this){
		if (cleanVector == null) cleanVector = new Vector();
		v = cleanVector;
		//
		// Cannot modify the hashtable while enumerating it.
		// So place keys in a Vector.
		//
		//System.out.println("Cleaning: "+table.size());
		for (Enumeration e = table.keys(); e.hasMoreElements();){
			WeakKey wk = (WeakKey)e.nextElement();
			if (wk.get() != null){ 
				//System.out.println("Got WK for: "+wk.get());
				continue;
			}
			v.add(wk);
		}
		int sz = v.size();
		if (sz != 0){
			for (int i = 0; i<sz; i++){
				WeakKey wk = (WeakKey)v.get(i);
				table.remove(wk);
				v.setElementAt(wk.data, i);
			}
			cleanVector = new Vector();
		}else
			v = null;
	}
	if (v != null) valuesRemoved(v);
}
/**
 * This is called if a clean() call finds values to remove. This call
 * is not synchronized with the WeakCache and by default it does nothing.
 * @param values the values that were removed.
 */
protected void valuesRemoved(Vector values)
{
	
}
/**
 * Given a value, find the key associated with that value if present.
 * @param value the value to look for.
 * @return the key for the value or null if not found or was garbage collected.
 */
public Object getKeyForValue(Object value)
{
	for (Iterator it = table.entrySet().iterator(); it.hasNext();){
		Map.Entry e = (Map.Entry)it.next();
		if (e.getValue().equals(value)){
			WeakKey wk = (WeakKey)e.getKey();
			return wk.get();
		}
	}
	return null;
}
/*
//-------------------------------------------------------------------
int indexOf(Object key)
//-------------------------------------------------------------------
{
	if (key == null) return -1;
	for (int i = 0; i<refs.size(); i++){
		WeakReference wr = (WeakReference)refs.get(i);
		if (wr == key) return i;
		Object g = wr.get();
		if (g == key) return i;
		if (g == null){
			refs.removeElementAt(i);
			data.removeElementAt(i);
			i--;
		}
	}
	return -1;
}
*/
//===================================================================
public void put(Object key, Object data)
//===================================================================
{
	if (key == null || data == null) throw new NullPointerException();
	table.put(new WeakKey(key,data,rq),data);
	/*
	if (key == null) throw new NullPointerException();
	int where = indexOf(key);
	if (where != -1) this.data.setElementAt(data,where);
	else {
		this.data.add(data);
		refs.add(key instanceof WeakReference ? key : new WeakReference(key));
	}
	*/
}
//===================================================================
public Object get(Object key)
//===================================================================
{
	return table.get(new WeakKey(key));
	/*
	if (key == null) return null;
	int where = indexOf(key);
	return where == -1 ? null : data.get(where);
	*/
}
public void remove(Object key)
{
	table.remove(new WeakKey(key));
}
/**
 * Each object returned by the returned Iterator will be a WeakReference.
 * @return an Iterator that iterates through the WeakReferences.
 */
/*
//===================================================================
public Iterator keys()
//===================================================================
{
	return refs.iterator();
}
*/
//##################################################################
}
//##################################################################

class Cleaner extends Thread {
	Cleaner()
	{
		super("WeakCache Cleaner Thread");
	}
	public void run(){
		Vector v = new Vector();
		while(true){
			WeakCache.caches.getRefs(v);
			int s = v.size();
			for (int i = 0; i<s; i++){
				WeakCache wc = (WeakCache)v.get(i);
				if (wc == null) continue;
				v.setElementAt(null, i);
				boolean gotOne = false;
				while(wc.rq.poll() != null)
					gotOne = true;
				if (gotOne) wc.clean();
				wc = null;
			}
			v.clear();
			try{
				sleep(1000);
			}catch(InterruptedException e){}
		}
	}
	/*
	WeakReference c;
	Cleaner(WeakCache wc)
	{
		c = new WeakReference(wc);
		start();
	}
	public void run(){
		int longTime = 10000, shortTime = 1000;
		Countdown cd = new Countdown();
		cd.start(longTime);
		int count = 0;
		while(true){
			ReferenceQueue rq = null;
			{
				WeakCache ww = (WeakCache)c.get();
				if (ww == null) return;
				rq = ww.rq;
				shortTime = ww.cleanDelay;
			}
			if (shortTime < 0 || shortTime > 1000) shortTime = 1000;
			try{
				long left = cd.timeRemaining();
				if (left == 0){
					if (count != 0) {
						WeakCache ww = (WeakCache)c.get();
						if (ww == null) continue; //Will exit at the top.
						ww.clean();
					}
					count = 0;
					cd.start(longTime);
					left = longTime;
				}
				Object got = rq.remove(left);
				if (got != null){
					count++;
					left = cd.timeRemaining();
					if (left > shortTime) cd.start(shortTime);
				}
			}catch(InterruptedException e){}
			
		}
	}
	*/
}
class WeakKey extends WeakReference{
	int myHashcode;
	Object data;
	public boolean equals(Object other)
	{
		return other == this || ((WeakKey)other).get() == get();
		//return other == get() || other == this;
	}
	public WeakKey(Object key)
	{
		super(key);
		myHashcode = key == null ? 0 : System.identityHashCode(key);//.hashCode();
	}
	public WeakKey(Object key, Object data, ReferenceQueue rq)
	{
		super(key,rq);
		myHashcode = key == null ? 0 : System.identityHashCode(key);//.hashCode();
		this.data = data;
	}
	public int hashCode()
	{
		return myHashcode;
	}
}
